#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cmath>

//------------------------------------------------------------------------------
// GELU approximation function
__device__ float gelu_approx(float x) {
    const float kAlpha = 0.044715f;
    const float kBeta  = 0.7978845608f; // sqrt(2/M_PI)
    float inner = kBeta * (x + kAlpha * x * x * x);
    return x * 0.5f * (1.0f + tanhf(inner));
}

//------------------------------------------------------------------------------
// Fused kernel with improved memory coalescing
// Each block processes one row from the input 'x'.
// To ensure coalesced global memory accesses, we compute row pointers 
// so that threads in a warp access consecutive memory locations when reading
// from 'x' and writing to 'out'.
__global__ void fused_forward_coalesced_kernel(
    const float* __restrict__ x,            // Input: shape [batch_size, in_features]
    const float* __restrict__ weight_sum,     // Precomputed: shape [in_features]
    float constant,                           // Precomputed scalar constant: sum(bias - subtract)
    float* __restrict__ out,                  // Output: shape [batch_size, in_features]
    int batch_size,
    int in_features,
    int out_features                        // Used for normalization
) {
    int row = blockIdx.x;
    if (row >= batch_size) return;

    // Set up pointers for coalesced access, since rows are contiguous
    const float* row_x = x + row * in_features;
    float* row_out = out + row * in_features;

    // Use shared memory for dot product reduction
    extern __shared__ float sdata[]; // Size should be at least blockDim.x
    float sum = 0.0f;

    // Each thread processes a contiguous chunk of the row
    for (int j = threadIdx.x; j < in_features; j += blockDim.x) {
        // Global memory accesses here are coalesced as threads access consecutive elements
        sum += row_x[j] * weight_sum[j];
    }

    sdata[threadIdx.x] = sum;
    __syncthreads();

    // Perform parallel reduction in shared memory
    for (int stride = blockDim.x / 2; stride > 0; stride >>= 1) {
        if (threadIdx.x < stride) {
            sdata[threadIdx.x] += sdata[threadIdx.x + stride];
        }
        __syncthreads();
    }

    float pool_val = sdata[0];
    if (threadIdx.x == 0) {
        // Normalize using out_features and add the constant
        pool_val = (pool_val + constant) / static_cast<float>(out_features);
        // Apply GELU activation
        pool_val = gelu_approx(pool_val);
        sdata[0] = pool_val; // Broadcast to other threads
    }
    __syncthreads();
    pool_val = sdata[0];

    // Coalesced write: each thread writes consecutively to the output row
    for (int j = threadIdx.x; j < in_features; j += blockDim.x) {
        row_out[j] = row_x[j] + pool_val;
    }
}

//------------------------------------------------------------------------------
// Forward function: Precompute weight_sum and constant, then launch the fused kernel
// using one block per row. Global memory accesses are aligned to ensure coalescing.

torch::Tensor forward_cuda_coalesced(
    const torch::Tensor& x,
    const torch::Tensor& weight,
    const torch::Tensor& bias,
    const torch::Tensor& subtract
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");
    TORCH_CHECK(subtract.is_cuda(), "subtract must be a CUDA tensor");
    TORCH_CHECK(x.dim() == 2, "x must be 2D (batch_size x in_features)");
    TORCH_CHECK(weight.dim() == 2, "weight must be 2D (out_features x in_features)");
    TORCH_CHECK(bias.dim() == 1, "bias must be 1D (out_features)");
    TORCH_CHECK(subtract.dim() == 1, "subtract must be 1D (out_features)");

    int64_t batch_size  = x.size(0);
    int64_t in_features = x.size(1);
    int64_t out_features = weight.size(0);

    TORCH_CHECK(weight.size(1) == in_features, "weight.shape[1] must match x.shape[1]");
    TORCH_CHECK(bias.size(0) == out_features, "bias.shape[0] must match weight.shape[0]");
    TORCH_CHECK(subtract.size(0) == out_features, "subtract.shape[0] must match weight.shape[0]");

    auto x_contig = x.contiguous();
    auto weight_contig = weight.contiguous();
    auto bias_contig = bias.contiguous();
    auto subtract_contig = subtract.contiguous();

    // Precompute weight_sum: sum along the 0-th dimension of weight gives a vector of size [in_features]
    auto weight_sum = torch::sum(weight_contig, 0);

    // Precompute constant = sum(bias - subtract) as a scalar
    auto constant_tensor = torch::sum(bias_contig - subtract_contig);
    float constant = constant_tensor.item<float>();

    // Allocate output tensor with the same shape as x
    auto out = torch::empty({batch_size, in_features}, x.options());

    int threads = 256;          // Number of threads per block
    int blocks = batch_size;      // One block per row
    size_t shared_mem_bytes = threads * sizeof(float);

    fused_forward_coalesced_kernel<<<blocks, threads, shared_mem_bytes>>>(
        x_contig.data_ptr<float>(),
        weight_sum.data_ptr<float>(),
        constant,
        out.data_ptr<float>(),
        batch_size,
        in_features,
        out_features
    );

    return out;
}

//------------------------------------------------------------------------------
// PyBind11 module registration
PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward_cuda_coalesced, "Fused Forward Coalesced CUDA Kernel");
}
